

from abc import ABC, abstractmethod

from twisted.internet.task import LoopingCall
from twisted.python.failure import Failure

from nucypher.utilities.logging import Logger


class SimpleTask(ABC):
    """Simple Twisted Looping Call abstract base class."""
    INTERVAL = 60  # 60s default

    def __init__(self):
        self.log = Logger(self.__class__.__name__)
        self.__task = LoopingCall(self.run)

    @property
    def running(self) -> bool:
        """Determine whether the task is already running."""
        return self.__task.running

    def start(self, now: bool = False):
        """Start task."""
        if not self.running:
            d = self.__task.start(interval=self.INTERVAL, now=now)
            d.addErrback(self.handle_errors)

    def stop(self):
        """Stop task."""
        if self.running:
            self.__task.stop()

    @abstractmethod
    def run(self):
        """Task method that should be periodically run."""
        raise NotImplementedError

    @abstractmethod
    def handle_errors(self, *args, **kwargs):
        """Error callback for error handling during execution."""
        raise NotImplementedError

    @staticmethod
    def clean_traceback(failure: Failure) -> str:
        # FIXME: Amazing.
        cleaned_traceback = failure.getTraceback().replace('{', '').replace('}', '')
        return cleaned_traceback
